@@ -15,7 +15,6 @@
 #
 
 import codecs
-utf_8_decode = codecs.utf_8_decode
 import datetime
 from io import UnsupportedOperation
 import os
@@ -23,8 +22,7 @@ import select
 import struct
 import zlib
 
-from extras import safe_hasattr, try_imports
-builtins = try_imports(['__builtin__', 'builtins'])
+from extras import safe_hasattr
 
 import subunit
 import subunit.iso8601 as iso8601
@@ -383,7 +381,7 @@ class ByteStreamToStreamResult(object):
                 raise ParseError(
                     'Bad checksum - calculated (0x%x), stored (0x%x)'
                         % (crc, packet_crc))
-            if safe_hasattr(builtins, 'memoryview'):
+            if safe_hasattr(__builtins__, 'memoryview'):
                 body = memoryview(packet[-1])
             else:
                 body = packet[-1]
@@ -425,7 +423,6 @@ class ByteStreamToStreamResult(object):
                     raise ParseError('File content extends past end of packet: '
                         'claimed %d bytes, %d available' % (
                         content_length, len(file_bytes)))
-                pos += content_length
             else:
                 file_name = None
                 file_bytes = None
@@ -455,12 +452,7 @@ class ByteStreamToStreamResult(object):
             raise ParseError('UTF8 string at offset %d contains NUL byte' % (
                 pos-2,))
         try:
-            utf8, decoded_bytes = utf_8_decode(utf8_bytes)
-            if decoded_bytes != length:
-                raise ParseError("Invalid (partially decodable) string at "
-                    "offset %d, %d undecoded bytes" % (
-                    pos-2, length - decoded_bytes))
-            return utf8, length+pos
+            return utf8_bytes.decode('utf-8'), length+pos
         except UnicodeDecodeError:
             raise ParseError('UTF8 string at offset %d is not UTF8' % (pos-2,))
 
